/*
 * Video Capture Driver ( Video for Linux 1/2 )
 * for the Matrox Marvel G200 and Rainbow Runner-G series
 *
 * This module is an interface to the G100 and G200 video extension
 * registers.  
 *
 * Copyright (C) 1999  Ryan Drake <stiletto@mediaone.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/poll.h>
#include <asm/io.h>
#include <linux/pci.h>
#include <linux/interrupt.h>
#include <linux/i2c.h>
#include "mgavideo.h"


#define PKMOD "mgavideo: "
#define debug_msg(fmt,arg...) printk(KERN_DEBUG PKMOD "[debug] " fmt,##arg)
#define err_msg(fmt,arg...) printk(KERN_ERR PKMOD fmt,##arg)
#define info_msg(fmt,arg...) printk(KERN_INFO PKMOD fmt,##arg)


/* i2c registers */
#define PALWTADD        0x3c00
#define X_DATAREG       0x3c0a
#define XGENIOCTRL      0x2a
#define XGENIODATA      0x2b
#define SDA             0x10
#define SCL             0x20

#define I2C_DELAY       10


/* video registers */
#define VINCTL0         0x3e00
#define VINCTL1         0x3e04
#define VBIADDR0        0x3e08
#define VBIADDR1        0x3e0c
#define VINADDR0        0x3e10
#define VINADDR1        0x3e14
#define VINNEXTWIN      0x3e18
#define VINCTL          0x3e1c
#define VSTATUS         0x3e30
#define VICLEAR         0x3e34
#define VIEN            0x3e38

#define CODECCTL        0x3e40
#define CODECADDR       0x3e44
#define CODECHOSTPTR    0x3e48
#define CODECHARDPTR    0x3e4c
#define CODECLCODE      0x3e50

#define VINCTLx_MASK    0x00000fff
#define VBIADDRx_MASK   0x00fff200
#define VINADDRx_MASK   0x00fffff8


/* mga video status */
#define VI_IVSYNC    0x0001 /* input vsync interrupt */
#define VI_CMDCMPL   0x0002 /* command complete interrupt */
#define VI_BLVL      0x0004 /* buffer level interrupt */
#define VI_EOI       0x0008 /* codec end of image interrupt */
#define VS_FIELD     0x0100 /* field 0 or 1 detected */
#define VS_VIDEO     0x0200 /* indicates video was captured */
#define VS_RAWVBI    0x0400 /* indicates raw vbi was captured */
#define VS_SLICEDVBI 0x0800 /* indicates sliced vbi was captured */


/****************************************************************************
* raw register access : these routines directly interact with the mga's
*                       control aperature.  must not be called until after
*                       the board's pci memory has been mapped.
****************************************************************************/
static u32 _mgaread( struct mga_dev* mga, u32 reg )
{
        return readl( mga->ctrl + reg );
}

static void _mgawrite( struct mga_dev* mga, u32 reg, u32 data )
{
        writel( data, mga->ctrl + reg );
}


/****************************************************************************
* mga i2c bus interface
****************************************************************************/
static void _mgai2c_attach( struct i2c_bus* bus, int id )
{
        struct mga_dev* mga = (struct mga_dev*)bus->data;

        debug_msg( "i2c device attach: %d\n", id );

        switch( id )
        {
        case I2C_DRIVERID_MSP3400:
                mga->hasaudio = 1;
                break;

        case I2C_DRIVERID_TUNER:
                mga->hastuner = 1;
                break;

        case I2C_DRIVERID_KS0127:
                mga->hasdecoder = 1;
                break;
        }
}

static void _mgai2c_detach( struct i2c_bus* bus, int id )
{
        struct mga_dev* mga = (struct mga_dev*)bus->data;

        debug_msg( "i2c device detach: %d\n", id );

        switch( id )
        {
        case I2C_DRIVERID_MSP3400:
                mga->hasaudio = 0;
                break;

        case I2C_DRIVERID_TUNER:
                mga->hastuner = 0;
                break;

        case I2C_DRIVERID_KS0127:
                mga->hasdecoder = 0;
                break;
        }
}

static void _mgai2c_setlines( struct i2c_bus* bus, int ctrl, int data )
{
        struct mga_dev* mga = (struct mga_dev*)bus->data;

        if( ctrl )
        {
        writeb( XGENIOCTRL, mga->ctrl + PALWTADD );
        writeb( readb( mga->ctrl + X_DATAREG ) & ~SCL, mga->ctrl + X_DATAREG );
        }
        else
        {
        writeb( XGENIODATA, mga->ctrl + PALWTADD );
        writeb( readb( mga->ctrl + X_DATAREG ) & ~SCL, mga->ctrl + X_DATAREG );
        writeb( XGENIOCTRL, mga->ctrl + PALWTADD );
        writeb( readb( mga->ctrl + X_DATAREG ) | SCL, mga->ctrl + X_DATAREG );
        }

        if( data )
        {
        writeb( XGENIOCTRL, mga->ctrl + PALWTADD );
        writeb( readb( mga->ctrl + X_DATAREG ) & ~SDA, mga->ctrl + X_DATAREG );
        }
        else
        {
        writeb( XGENIODATA, mga->ctrl + PALWTADD );
        writeb( readb( mga->ctrl + X_DATAREG ) & ~SDA, mga->ctrl + X_DATAREG );
        writeb( XGENIOCTRL, mga->ctrl + PALWTADD );
        writeb( readb( mga->ctrl + X_DATAREG ) | SDA, mga->ctrl + X_DATAREG );
        }

        udelay( I2C_DELAY );
}

static int _mgai2c_getdataline( struct i2c_bus* bus )
{
        struct mga_dev* mga = (struct mga_dev*)bus->data;

        writeb( XGENIOCTRL, mga->ctrl + PALWTADD );
        writeb( readb( mga->ctrl + X_DATAREG ) & ~SDA, mga->ctrl + X_DATAREG );
        writeb( XGENIODATA, mga->ctrl + PALWTADD );
        return ( readb( mga->ctrl + X_DATAREG ) & SDA )?1:0;
}


/****************************************************************************
* mgavideo api : public interface to the mga's video subsystem
****************************************************************************/
/* component device control */
void mgavideo_tuner( struct mga_dev* mga, unsigned int cmd, void* arg )
{
        if( mga->hastuner )
                i2c_control_device( &(mga->i2c), I2C_DRIVERID_TUNER, cmd, arg );
}

void mgavideo_audio( struct mga_dev* mga, unsigned int cmd, void* arg )
{
        if( mga->hasaudio )
                i2c_control_device( &(mga->i2c), I2C_DRIVERID_MSP3400, cmd, arg );
}

void mgavideo_decoder( struct mga_dev* mga, unsigned int cmd, void* arg )
{
        if( mga->hasdecoder )
                i2c_control_device( &(mga->i2c), I2C_DRIVERID_KS0127, cmd, arg );
}

/* video grabber control */
void mgavideo_grabber_enable( struct mga_dev* mga, int en )
{
        debug_msg( "%sable grabber\n", en?"en":"dis" );
        _mgawrite( mga, VINCTL, en?0x0101ab01:0x0101ab00 );
}

/* video interrupt control */
void mgavideo_ivsync_enable( struct mga_dev* mga, int en )
{ 
        u32 istat = _mgaread( mga, VIEN );
        debug_msg( "%sable interrupt\n", en?"en":"dis" );
        _mgawrite( mga, VIEN, en?(istat|VI_IVSYNC):(istat&~VI_IVSYNC) );
}

/* check whether a video interrupt is pending */
int mgavideo_ivsync_pending( struct mga_dev* mga )
{
        return _mgaread( mga, VSTATUS ) & VI_IVSYNC;
}

/* clear the video interrupt pending flag */
void mgavideo_ivsync_clear( struct mga_dev* mga )
{
        _mgawrite( mga, VICLEAR, VI_IVSYNC );
}

/* set the next window for the grabber to get */
void mgavideo_nextwin( struct mga_dev* mga, int window )
{
        _mgawrite( mga, VINNEXTWIN, window?1:0 );
}

/* set vbi capture mode */
void mgavideo_set_mode0( struct mga_dev* mga, int vin, int vbi, int pitch )
{
        debug_msg( "window: 0  vin: %d  vbi: %d  pitch: %d\n", vin, vbi, pitch );
        _mgawrite( mga, VINCTL0, (vin | vbi | pitch << 3) & VINCTLx_MASK );
}

void mgavideo_set_mode1( struct mga_dev* mga, int vin, int vbi, int pitch )
{
        debug_msg( "window: 1  vin: %d  vbi: %d  pitch: %d\n", vin, vbi, pitch );
        _mgawrite( mga, VINCTL1, (vin | vbi | pitch << 3) & VINCTLx_MASK );
}

/* set vbi start address */
void mgavideo_set_vbiaddr0( struct mga_dev* mga, u32 addr )
{
        debug_msg( "vbi window: 0  address: %x\n", addr );
        _mgawrite( mga, VBIADDR0, addr & VBIADDRx_MASK );
}

void mgavideo_set_vbiaddr1( struct mga_dev* mga, u32 addr )
{
        debug_msg( "vbi window: 1  address: %x\n", addr );
        _mgawrite( mga, VBIADDR1, addr & VBIADDRx_MASK );
}

/* set video start address */
void mgavideo_set_vinaddr0( struct mga_dev* mga, u32 addr )
{
        debug_msg( "video window: 0  address: %x\n", addr );
        _mgawrite( mga, VINADDR0, addr & VINADDRx_MASK );
}

void mgavideo_set_vinaddr1( struct mga_dev* mga, u32 addr )
{
        debug_msg( "video window: 1  address: %x\n", addr );
        _mgawrite( mga, VINADDR1, addr & VINADDRx_MASK );
}

/****************************************************************************
* initialization
****************************************************************************/
#define NBOARDS 4
struct mga_dev mgavideo_dev[NBOARDS];

/* find all mga adapters */
struct mga_dev* mgavideo_get( void )
{
        int i;
        for( i=0; i<NBOARDS; i++ )
        {
                if( !mgavideo_dev[i].inuse )
                {
                        mgavideo_dev[i].inuse++;
                        MOD_INC_USE_COUNT;
                        return &mgavideo_dev[i];
                }
        }

        return NULL;
}

/* release a mga adapter that is not needed anymore */
void mgavideo_release( struct mga_dev* mga )
{
        if( !mga )
                return;

        if( mga->ctrl )
        {
                mgavideo_grabber_enable( mga, 0 );
                mgavideo_ivsync_enable( mga, 0 );
                mga->inuse--;
                MOD_DEC_USE_COUNT;
        }
}


/****************************************************************************
* linux kernel module api
****************************************************************************/
#ifdef MODULE
int init_module( void )
#else
int mgavideo_init( void )
#endif
{
        int found = 0;
        struct pci_dev* dev;
        struct mga_dev* mga;
        u32 addr = 0;

        debug_msg( "Looking for boards...\n" );

        /* find the next MGA G100 or G200 */
        for( dev = pci_devices; dev != NULL; dev = dev->next )
        {
                if (dev->class != PCI_CLASS_NOT_DEFINED_VGA &&
                  ((dev->class) >> 16 != PCI_BASE_CLASS_DISPLAY))
                        continue;

                if (PCI_FUNC(dev->devfn) != 0)
                        continue;

                /* ignore non-mga adapters */
                if (dev->vendor != PCI_VENDOR_ID_MATROX)
                        continue;

                /* ignore non-gseries adapters */
                if (dev->device != PCI_DEVICE_ID_MATROX_G200_PCI &&
                    dev->device != PCI_DEVICE_ID_MATROX_G200_AGP &&
                    dev->device != PCI_DEVICE_ID_MATROX_G100_MM &&
                    dev->device != PCI_DEVICE_ID_MATROX_G100_AGP)
                        continue;


                /* the device is cool. set up a context for it */
                mga = &mgavideo_dev[found];
                memset( mga, 0, sizeof( *mga ) );

                mga->device = dev->device;
                mga->inuse = 0;

                /* read and map the control aperature */
                addr = dev->base_address[1] & PCI_BASE_ADDRESS_MEM_MASK;
                mga->ctrl = (u32)ioremap( addr, 0x4000 );

                /* read the device's irq */
                mga->irq = dev->irq;

                /* reset the device */
                _mgawrite( mga, VINCTL, 0x0101ab00 );


                /* setup the i2c */
                sprintf( mga->i2c.name, "mga%d", found );
                mga->i2c.id = I2C_BUSID_MGA;
                mga->i2c.data = mga;
            #if LINUX_VERSION_CODE >= 0x020100
                mga->i2c.bus_lock = SPIN_LOCK_UNLOCKED;
            #endif
                mga->i2c.attach_inform = _mgai2c_attach;
                mga->i2c.detach_inform = _mgai2c_detach;
                mga->i2c.i2c_setlines = _mgai2c_setlines;
                mga->i2c.i2c_getdataline = _mgai2c_getdataline;
                mga->i2c.i2c_read = NULL;
                mga->i2c.i2c_write = NULL;
                i2c_register_bus( &mga->i2c );


                /* okay everything looks good */
                info_msg( "Found %s: device 0x%x irq %i\n", mga->i2c.name, mga->device, mga->irq );
                
                if( ++found >= NBOARDS )
                        break;
        }

        if( !found )
                return -ENODEV;

        return 0;
}

#ifdef MODULE
void cleanup_module( void )
{
        int i;
        struct mga_dev* mga;

        for( i=0; i<NBOARDS; i++ )
        {
                mga = &mgavideo_dev[i];
                if( mga->ctrl )
                {
                        i2c_unregister_bus( &mga->i2c );
                        iounmap( (void*)mga->ctrl );
                }
        }
}
#endif
